//
//  IsingWindowCtl.m
//  IsingCocoa
//
//  Created by Cameron Hayne on Tue May 21 2002.
//  Copyright Hayne of Tintagel 2005
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#import "IsingWindowCtl.h"
#import "IsingToolbar.h"
#import "IsingStats.h"
#import "IsingModel.h"
#import "IsingSpinView.h"
#import "IsingViewCtl.h"
#import "SizeConfigPanelCtl.h"
#import "SpinsConfigPanelCtl.h"
#import "BlockSpinsParamsPanelCtl.h"
#import "WindowSizing.h"
#import "Util.h"

// private methods
@interface IsingWindowCtl (Private)
- (void)startViewUpdateTimer;
- (void)stopViewUpdateTimer;
- (void)updateViews:(NSTimer *)aTimer;
@end

@implementation IsingWindowCtl

- (id)init
{
    self = [super initWithWindowNibName: @"IsingWindow"];
    if (self == nil)
        return nil;

    //NSLog(@"IsingWindowCtl.init");
    
    // We use 'setShouldCloseDocument' so that closing the main IsingWindow
    // will close the document and hence close all auxiliary windows
    [self setShouldCloseDocument: YES];
    
    // the viewController holds the set of spins that is on display
    // (a snapshot of the spins from the model)
    // and it tells all the views when an update is necessary
    viewController = [[IsingViewCtl alloc] init];

    needsViewUpdate = NO;
    return self;
}

- (void)dealloc
{
    //NSLog(@"IsingWindowCtl.dealloc");
    [viewController release];
    [viewUpdateTimer release];
    [super dealloc];
}

- (void)windowWillLoad
{
    [Util checkAppKitVersion]; // aborts if not Panther or later
}

- (void)windowDidLoad
{
    //NSLog(@"IsingWindowCtl.windowDidLoad");
    [super windowDidLoad];
    [[self window] useOptimizedDrawing: YES]; // does this make a difference?

    [self addIsingView: spinView];

    [self setupToolbar]; // see IsingToolbar.m
                         // note that we do this before the initModelSize
                         // since the toolbar changes the size of the window
    
    // the field slider is visible in the NIB,
    // so we need to toggle it since we want it to default to off
    [self toggleMagneticFieldAction: nil];
    
    // set up the model for the current window size
    NSSize boundsSize = [spinView bounds].size;
    IsingSize size = makeIsingSize(boundsSize.width, boundsSize.height);
    [self initModelSize: size];
    
    [self startViewUpdateTimer];
    
    // start the calculation thread
    [self pauseButtonAction: nil];
    info.startTime = [NSDate timeIntervalSinceReferenceDate];
}

- (void)windowWillClose:(NSNotification *)aNotification
{
    //NSLog(@"IsingWindowCtl.windowWillClose");
    [model stopUpdate];
    [self stopViewUpdateTimer];
}

// closing the last window quits the app ?
//- (BOOL)applicationShouldTerminateAfterLastWindowClosed:(NSApplication *)theApplication
//{
//    return YES;
//}

#pragma mark Resizing methods

// called from applicationDidFinishLaunching and closeSpecifySizeSheet
- (void)initModelSize:(IsingSize)size
{    
    // ask the model to change its size to match the given size
    BOOL modelReinit;
    IsingSize mSize = [model requestSize: size
                          modelWasReinit: &modelReinit];
    if (mSize.width != size.width || mSize.height != size.height)
    {
        // something went wrong with the size constraints
        NSLog(@"initModelSize: size constraints failed: size %d %d modelSize %d %d",
              size.width, size.height, mSize.width, mSize.height);
    }
    [viewController updateIsingViewSpinsSize: mSize];
    [self synchronizeWindowTitleWithDocumentName];
    
    if (modelReinit == YES)
    {
        // reset the bookkeeping variables
        [self resetStats];
        lastSweepDisplayed = -1;
    }
}

// The controller gets notified of window resizes since it is the window's delegate
- (NSSize)windowWillResize:(NSWindow *)window toSize:(NSSize)proposedFrameSize
{
    // calculate what size the contentRect of the window would be
    NSRect currentFrame = [window frame];
    NSRect proposedFrame =
        NSMakeRect(NSMinX(currentFrame),
                   NSMinY(currentFrame)
                   + (NSHeight(currentFrame) - proposedFrameSize.height),
                   proposedFrameSize.width,
                   proposedFrameSize.height);
    NSRect currentContentRect  = [NSWindow contentRectForFrameRect: currentFrame
                                                         styleMask: [window styleMask]];
    NSRect proposedContentRect = [NSWindow contentRectForFrameRect: proposedFrame
                                                         styleMask: [window styleMask]];
    
    float deltaWidth  = NSWidth(proposedContentRect) - NSWidth(currentContentRect);
    float deltaHeight = NSHeight(proposedContentRect) - NSHeight(currentContentRect);
    //NSLog(@"windowWillResize: deltaWidth = %f deltaHeight = %f",
    //                          deltaWidth, deltaHeight);
    
    // we assume that all of the changes will go into the spinView
    // and we want to maintain the current aspect ratio of the spinView
    NSSize currentViewSize = [spinView frame].size;
    float ratio = currentViewSize.width / currentViewSize.height;
    float expectedDeltaWidth = (deltaWidth + deltaHeight * ratio) / 2.0;
    float expectedDeltaHeight = (deltaWidth / ratio + deltaHeight) / 2.0;
    
    NSSize maxUsableSize = [window maxUsableViewSize: spinView];
    if (currentViewSize.height + expectedDeltaHeight > maxUsableSize.height)
    {
        expectedDeltaHeight = maxUsableSize.height - currentViewSize.height;
        expectedDeltaWidth = maxUsableSize.height * ratio - currentViewSize.width;
    }
    else if (currentViewSize.width + expectedDeltaWidth > maxUsableSize.width)
    {
        expectedDeltaWidth = maxUsableSize.width - currentViewSize.width;
        expectedDeltaHeight = maxUsableSize.width / ratio - currentViewSize.height;
    }
    
    float widthCorrection  = expectedDeltaWidth - deltaWidth;
    float heightCorrection = expectedDeltaHeight - deltaHeight;
    //NSLog(@"windowWillResize: widthCorrection = %f heightCorrection = %f",
    //                          widthCorrection, heightCorrection);
    
    // calculate what size this implies for the window frame
    float acceptableContentWidth  = NSWidth(proposedContentRect) + widthCorrection;
    float acceptableContentHeight = NSHeight(proposedContentRect) + heightCorrection;
    NSRect acceptableContentRect =
        NSMakeRect(NSMinX(proposedContentRect),
                   NSMinY(proposedContentRect) - heightCorrection,
                   acceptableContentWidth,
                   acceptableContentHeight);
    NSRect acceptableFrame = [NSWindow frameRectForContentRect:acceptableContentRect
                                                     styleMask:[window styleMask]];
    return acceptableFrame.size;
}

- (NSString *)windowTitleForDocumentDisplayName:(NSString *)displayName
{
    IsingSize size = [model size];
    NSString *title = [NSString stringWithFormat:@"Ising %d x %d (%@)",
                                        size.width, size.height, displayName];
    return title;
}

#pragma mark ViewUpdate methods

// the viewController keeps track of all of the IsingView's that need to be updated
- (void)addIsingView:(id <IsingView>)aView
{
    [viewController addView: aView];
}

// this method is invoked via the timer
- (void)updateViews:(NSTimer *)aTimer
{
    // The following print statement shows that the interval between timer invocations
    // is not always constant - there are sometimes large delays
    //printf("%g\n", [NSDate timeIntervalSinceReferenceDate] - info.startTime);
    
    BOOL updateDesired = needsViewUpdate
                         || ([model numSweeps] > lastSweepDisplayed);
    if (updateDesired)
    {
        // Tell the views to redraw themselves with the new spins
        [viewController updateIsingViewSpins: [model lockSpinsForDisplay]];
        lastSweepDisplayed = [model numSweeps];
        needsViewUpdate = NO;
        [model unlockSpinsForDisplay];
        
        // Calculate the update rates and display them in the drawer
        [self updateStatistics];
    }
}

- (void)startViewUpdateTimer
{
    if (viewUpdateTimer != nil)
        return;
    
    int fpsMax = 30;
    NSTimeInterval viewUpdatePeriod = 1.0 / fpsMax;
    
    viewUpdateTimer = [[NSTimer scheduledTimerWithTimeInterval: viewUpdatePeriod
                                                        target: self
                                                      selector: @selector(updateViews:)
                                                      userInfo: nil
                                                       repeats: YES] retain];
    
    // We need the following line so that the view updates continue
    // when the user is dragging a slider
    [[NSRunLoop currentRunLoop] addTimer: viewUpdateTimer
                                 forMode: NSEventTrackingRunLoopMode];
}

- (void)stopViewUpdateTimer
{
    if (viewUpdateTimer == nil)
        return;
    
    [viewUpdateTimer invalidate];
    [viewUpdateTimer release];
    viewUpdateTimer = nil;
}

#pragma mark Menu updates

- (BOOL)validateMenuItem:(NSMenuItem *)menuItem
{
    //NSLog(@"IsingWindowCtl.validateMenuItem %@", [menuItem title]);
    SEL action = [menuItem action];
    if (action == @selector(statsDrawerAction:))
    {
        NSDrawerState state = [statsDrawer state];
        if (state == NSDrawerOpenState || state == NSDrawerOpeningState)
        {
            [menuItem setTitle:@"Hide Statistics"];
        }
        else
        {
            [menuItem setTitle:@"Show Statistics"];
        }
    }
    else if (action == @selector(toggleMagneticFieldAction:))
    {
        [menuItem setTitle: [model fieldIsPresent] ? @"Disable Magnetic Field"
                                                   : @"Enable Magnetic Field"];
    }
    else if (action == @selector(pauseButtonAction:))
    {
        [menuItem setTitle: [model updateIsActive] ? @"Pause"
                                                   : @"Resume"];
    }
        
    return YES; // default is for all other menuItems to be enabled
}

#pragma mark Actions

// action to open & close the drawer
- (IBAction)statsDrawerAction:(id)sender
{
    NSDrawerState state = [statsDrawer state];
    if (state == NSDrawerOpenState || state == NSDrawerOpeningState)
    {
        [statsDrawer close];
    }
    else
    {
        [statsDrawer openOnEdge: NSMaxXEdge];
    }
}

- (IBAction)resetAvgFpsAction:(id)sender
{
    [self resetFpsBookkeeping];
}

- (IBAction)modelUpdateRateAction:(id)sender
{
    float sliderValue = [sender floatValue];
    float rate = pow(10.0, sliderValue); // slider is logarithmic
    //NSLog(@"slidervalue: %g rate: %g", sliderValue, rate);
    [model setMaxUpdateRate: rate];
}

- (IBAction)pauseButtonAction:(id)sender
{
    // toggle on & off updating of model
    if ([model updateIsActive])
    {
        [model stopUpdate];
    }
    else
    {
        [self resetStatsUpdateTime];
        [model startUpdate];
    }
    
    [self updateToolbarItems];
}

- (IBAction)toggleMagneticFieldAction:(id)sender
{
    if ([model fieldIsPresent])
    {        
        // disable the magnetic field and set it to zero for next time
        [model setField: 0.0];
        [model setFieldIsPresent: NO];
    }
    else
    {        
        [model setFieldIsPresent: YES];
    }
    
    [self updateToolbarItems];
    [[self window] toggleViewAndResize: fieldControl];
}

- (IBAction)setSizeAction:(id)sender
{
    [[[SizeConfigPanelCtl alloc] initForWindowCtl: self] autorelease];
}

- (IBAction)setSpinsAction:(id)sender
{
    [[[SpinsConfigPanelCtl alloc] initForWindowCtl: self] autorelease];
}

- (IBAction)blockSpinsAction:(id)sender
{
    [[[BlockSpinsParamsPanelCtl alloc] initForWindowCtl: self] autorelease];
}

#pragma mark Accessors

- (IsingModel *)model
{
    return model;
}

- (IsingSpinView *)spinView
{
    return spinView;
}

- (void)setNeedsViewUpdate:(BOOL)flag
{
    needsViewUpdate = flag;
}

@end
