//
//  ising.c
//  IsingCocoa
//
//  Created by Cameron Hayne on August 4, 1991.
//  (Modified from code written when I was at Visual Edge Software in 1991)
//  Copyright Hayne of Tintagel 2005
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <string.h>  // for memcpy

#include "ising.h"
#include "misc.h"
#include "randnum.h"

/* We will store each row of n_cols spins in n_wcols spin-words
   where n_wcols = n_cols / WORD_SIZE + ((n_cols % WORD_SIZE) ? 1 : 0).
   The ((n_cols % WORD_SIZE) ? 1 : 0) part is to add an extra spin_word
   in the case that n_cols is not a multiple of WORD_SIZE.
   (In this case, there are n_extra = (WORD_SIZE - n_cols % WORD_SIZE) spins
    in the last spin_word of each row. These extra spins will participate
    in the calculation but will not be displayed or included in statistics.)
   The spin-word arrays are one-dimensional arrays of length
   (n_rows + 2) * (n_wcols + 2) so we can think of them as two-dimensional
   arrays of size n_rows by n_wcols with a padding layer one word thick on
   left & right, and a padding layer one spin thick on top & bottom.
   The function of the padding layers is to supply periodic boundary conditions.

	BbbbbbB		The case shown is: n_rows = 4, n_wcols = 5
	bsssssb		The total number of words (including padding)
	bsssssb		is (n_rows+2)*(n_wcols+2) = 6 * 7 = 42
	bsssssb		We will ignore the corner spin-words 'B'.
	bsssssb		Each spin-word 'b' will be set to match the
	BbbbbbB		spin-word 's' that is on the far side from it
			in order to enforce periodic boundary conditions
   Note that the index of the top-left 's' spin-word is (n_wcols + 3)
*/

/***************************************************************************
Precomputed Constants
---------------------------------------------------------------------------*/
static SpinWord     AllBitsSet = ~0;
// HighBits[n - 1] is a mask yielding the high 'n' bits
static SpinWord     HighBits[WORD_SIZE];
// LowBits[n - 1] is a mask yielding the low 'n' bits
static SpinWord     LowBits[WORD_SIZE];
// BitsInChar[n] is the number of '1' bits in the 8-bit value 'n'
static unsigned int BitsInChar[256];

/***************************************************************************
NAME:		precompute_ising_constants()
INPUT:		---
OUTPUT:		---
RETURN:		---
DESCRIPTION:	Precomputes some constants.
EXT REFERENCES: ---
EXT EFFECTS:	Initializes the above precomputed constants
CREATION:	Apr 14/05

---------------------------------------------------------------------------*/
static void precompute_ising_constants()
{
    unsigned int i;

    for (i = 0; i < WORD_SIZE; i++)
    {
        HighBits[i] = (AllBitsSet << (WORD_SIZE - (i + 1)));
    }
    
    for (i = 0; i < WORD_SIZE; i++)
    {
        LowBits[i] = (AllBitsSet >> (WORD_SIZE - (i + 1)));
    }
 
    for (i = 0; i < 256; i++)
    {
        unsigned int num = i;
        unsigned int count = 0;
        while (num)
        {
            count += (num & 0x1u);
            num >>= 1 ;
        }
        BitsInChar[i] = count;
    }

//#define VERIFY_PRECOMPUTED
#ifdef VERIFY_PRECOMPUTED
    // verify that all is correct by looking at some values
    static int count_up_spins(SpinWord spin, int num_bits);
    for (i = 0; i < WORD_SIZE; i++)
    {
        printf("HighBits[%d] = %s\n", i, spinWordToString(HighBits[i]));
        printf("num set bits: %d\n", count_up_spins(HighBits[i], WORD_SIZE));
        printf("LowBits[%d] = %s\n", i, spinWordToString(LowBits[i]));
        printf("num set bits: %d\n", count_up_spins(LowBits[i], WORD_SIZE));
        printf("num set bits in low 7: %d\n", count_up_spins(LowBits[i], 7));
        printf("num set bits in low 14: %d\n", count_up_spins(LowBits[i], 14));
        SpinWord r = random();
        printf("r = %s\n", spinWordToString(r));
        printf("num set bits in low 9: %d\n", count_up_spins(r, 9));
    }
#endif
}

/***************************************************************************
NAME:		init_ising()
INPUT:		---
OUTPUT:		---
RETURN:		void
DESCRIPTION:	Does various initialization.
                NB: this function must be called before using any other
                function in this file.
                It is safe to call this function more than once
                - the subsequent calls do nothing.
EXT REFERENCES: ---
EXT EFFECTS:	---
CREATION:	May 7/05

---------------------------------------------------------------------------*/
void init_ising()
{
    static int init = 0;
    if ( ! init )
    {
        INIT_RAND_NUM(time(0));
        precompute_ising_constants();
        init = 1;
    }
}

/***************************************************************************
NAME:		spinWordToString( spin )
INPUT:		SpinWord	spin		: a spin word
OUTPUT:		---
RETURN:		char *                          : string representation of spin
DESCRIPTION:	Returns a string of 0's & 1's representing 'spin'
                This is mostly intended for debugging.
                Note: The string is in a static internal buffer
                      and should not be freed.)
EXT REFERENCES: ---
EXT EFFECTS:	---
CREATION:	Apr 14/05

---------------------------------------------------------------------------*/
char *spinWordToString( SpinWord spin )
{
    static char str[WORD_SIZE + 1];
    int i;
    
    SpinWord mask = HIGH_BIT;
    for (i = 0; i < WORD_SIZE; i++)
    {
        str[i] = (spin & mask) ? '1' : '0';
        mask >>= 1;
    }
    str[WORD_SIZE] = '\0';
    
    return str;
}

/***************************************************************************
NAME:		left_rotate_spin_word( spin )
INPUT:		SpinWord	spin		: a spin word
OUTPUT:		---
RETURN:		SpinWord			: left-rotated spin word
DESCRIPTION:	Returns a spin-word which is the result of left-rotating
		the bits of 'spin'.
EXT REFERENCES: ---
EXT EFFECTS:	---
CREATION:	Dec 4/91

---------------------------------------------------------------------------*/
static	SpinWord	left_rotate_spin_word( SpinWord spin )
{
    if ( spin & HIGH_BIT )
    {
        /* The high bit is set,
        so we need to set the low bit after shifting */ 
        
        return ( (spin << 1) | 1 );
    }
    else
    {
        /* The high bit is not set,
        so we need only shift - the low bit will get reset */ 
        
        return ( spin << 1 );
    }
}

/***************************************************************************
NAME:		right_rotate_spin_word( spin )
INPUT:		SpinWord	spin		: a spin word
OUTPUT:		---
RETURN:		SpinWord			: right-rotated spin word
DESCRIPTION:	Returns a spin-word which is the result of right-rotating
		the bits of 'spin'.
EXT REFERENCES: ---
EXT EFFECTS:	---
CREATION:	Dec 4/91

---------------------------------------------------------------------------*/
static	SpinWord	right_rotate_spin_word( SpinWord spin )
{
    if ( spin & 1 )
    {
        /* The low bit is set,
        so we need to set the high bit after shifting */ 
        
        return ( (spin >> 1) | HIGH_BIT );
    }
    else
    {
        /* The low bit is not set,
        so we need only shift - the high bit will get reset */ 
        
        return ( spin >> 1 );
    }
}

/***************************************************************************
NAME:		sumSpinWords( spinWords, numSpinWords, sumDigits, numSumDigits )
INPUT:		SpinWord    *spinWords      : array of spin-words to be summed
                int         numSpinWords    : number of spin-words in 'in' array
OUTPUT:		SpinWord    *sumDigits      : array giving binary digits of sum
                int         numSumDigits    : number of spin-words in 'out' array
RETURN:		void
DESCRIPTION:	Sums the spin-words using bitwise operations. The 'sumDigits' array
                will contain the binary digits of the sum.
                E.g. sumDigits[0] will contain the "ones" of the sums,
                     sumDigits[1] will contain the "twos" of the sums,
                     sumDigits[2] will contain the "fours" of the sums, etc
EXT REFERENCES: ---
EXT EFFECTS:	---
CREATION:	Apr 26/05

---------------------------------------------------------------------------*/
void sumSpinWords(SpinWord *spinWords, int numSpinWords,
                  SpinWord *sumDigits, int numSumDigits)
{
    for (int j = 0; j < numSumDigits; j++)
    {
        sumDigits[j] = 0;
    }
    
    for (int i = 0; i < numSpinWords; i++)
    {
        SpinWord val = spinWords[i];
        SpinWord carry;
        for (int j = 0; j <= numSumDigits; j++)
        {
            carry = sumDigits[j] & val;
            sumDigits[j] ^= val;
            
            if (carry)
            {
                val = carry;
            }
            else
            {
                break;
            }
        }
    }
}

/***************************************************************************
NAME:		update_boundary_layer( spinWords, n_rows, n_wcols )
INPUT:		SpinWord	*spinWords	: array of spins
		int		n_rows		: # of rows of spins
		int		n_wcols		: # of columns of spin-words
                                                  (not including the boundary spins)
OUTPUT:		---
RETURN:		void
DESCRIPTION:	The given one-dimensional array of spins is treated as a
		2D array of 'n_rows' by 'n_wcols' with a layer of boundary
		spins on all sides.  This routine sets the boundary spins to
		match their opposite numbers in order to enforce periodic
		boundary conditions.
EXT REFERENCES: ---
EXT EFFECTS:	---
CREATION:	Aug 4/91

---------------------------------------------------------------------------*/
static	void	update_boundary_layer( SpinWord *spinWords,
                                       int n_rows, int n_wcols )
{
    SpinWord	*spin_ptr;
    int		row_size, row, wcol, offset;
    
    /* Rows are horizontal, columns are vertical */
    row_size = n_wcols + 2;
    
    /* top row */
    spin_ptr = &spinWords[1];
    offset = n_rows * row_size;
    for (wcol = 0; wcol < n_wcols; wcol++, spin_ptr++)
    {
        *spin_ptr = spin_ptr[offset];
    }
    
    /* sides */
    spin_ptr = &spinWords[row_size];
    for (row = 0; row < n_rows; row++)
    {
        /* left side */
        *spin_ptr = left_rotate_spin_word( spin_ptr[n_wcols] );
        spin_ptr += (n_wcols + 1);
        
        /* right side */
        *spin_ptr = right_rotate_spin_word( spin_ptr[-n_wcols] );
        ++spin_ptr;
    }
    
    /* bottom row */
    spin_ptr = &spinWords[(n_rows + 1) * row_size + 1];
    offset = n_rows * row_size;
    for (wcol = 0; wcol < n_wcols; wcol++, spin_ptr++)
    {
        *spin_ptr = spin_ptr[-offset];
    }
}

/***************************************************************************
NAME:		randomize_spin_word( spin_ptr )
INPUT:		---
OUTPUT:		SpinWord	*spin_ptr		: ptr to a spin word
RETURN:		void
DESCRIPTION:	Sets the bits of '*spin_ptr' to random values.
EXT REFERENCES: ---
EXT EFFECTS:	---
CREATION:	Dec 3/91

---------------------------------------------------------------------------*/
static	void	randomize_spin_word( SpinWord *spin_ptr )
{
    short           bit;
    unsigned int    mask = 1;
    
    *spin_ptr = 0;
    
    for (bit = 0; bit < WORD_SIZE; bit++, mask <<= 1)
    {
        if ( RAND_NUM() > HALF_MAX )
        {
            *spin_ptr |= mask;
        }
    }
}

/***************************************************************************
NAME:		initialize_spins( spins, init_state )
INPUT:		Spins         *spins		: ptr to spins structure
                InitState_t	init_state	: desired initial spin state
OUTPUT:		---
RETURN:		void
DESCRIPTION:	Initializes the spins to a configuration specified by 'init_state'.
EXT REFERENCES: ---
EXT EFFECTS:	---
CREATION:	Aug 4/91

---------------------------------------------------------------------------*/
void	initialize_spins( Spins *spins, IsingInitState init_state )
{
    int         n_rows = spins->numRows;
    int         n_cols = spins->numCols;
    SpinWord    *spinWords = spins->spinWords;

    int n_wcols = num_spinwords_per_row(n_cols);
    int	row;
    
    // start at the first real spin
    SpinWord *spin_ptr = &spinWords[n_wcols + 3];
    for (row = 0; row < n_rows; row++)
    {
        int	wcol;
        
        for (wcol = 0; wcol < n_wcols; wcol++)
        {
            switch ( init_state )
            {
                case AllSpinsDown:
                    *spin_ptr = 0;
                    break;

                case AllSpinsUp:
                    *spin_ptr = ~0;
                    break;

                case RandomSpins:
                default:
                    randomize_spin_word( spin_ptr );
                    break;
            }
            
            ++spin_ptr;
        }
        
        /* Skip over the right and left boundary layer spins */
        spin_ptr += 2;
    }
    
    update_boundary_layer( spinWords, n_rows, n_wcols );
}

/***************************************************************************
NAME:		update_spin_word( spin_ptr, row_size, boltz_up )
INPUT:		SpinWord	*spin_ptr	: ptr to spin word
		int		row_size	: # of words per row
                                                  (including the boundary words)
		BoltzFactor	boltz_up[5]	: probabilities of a "down" spin
						  given the # of "up" neighbours
OUTPUT:		---
RETURN:		void
DESCRIPTION:	Updates the given spin word.
EXT REFERENCES: ---
EXT EFFECTS:	---
CREATION:	Dec 6/91

---------------------------------------------------------------------------*/
static	void	update_spin_word( SpinWord *spin_ptr, int row_size,
			          BoltzFactor boltz_up[5] )
{
    SpinWord	spin0, spin1, spin2, spin3, sum0, sum1, sum2, carry, mask;
    short	bit;
    
    BoltzFactor boltz_up0 = boltz_up[0],
                boltz_up1 = boltz_up[1],
                boltz_up2 = boltz_up[2],
                boltz_up3 = boltz_up[3],
                boltz_up4 = boltz_up[4];  // optimization
    
    spin0 = spin_ptr[-1];
    spin1 = spin_ptr[1];
    spin2 = spin_ptr[-row_size];
    spin3 = spin_ptr[row_size];
    
    sum0 = spin0 ^ spin1;
    sum1 = spin0 & spin1;
    
    carry = sum0 & spin2;
    sum0 = sum0 ^ spin2;
    sum1 = sum1 ^ carry;
    
    carry = sum0 & spin3;
    sum0 = sum0 ^ spin3;
    
    sum2 = sum1 & carry;
    sum1 = sum1 ^ carry;

    #ifdef CHECK_CALC
    SpinWord spinWords[4] = {spin0, spin1, spin2, spin3};
    SpinWord sumDigits[3];
    sumSpinWords(spinWords, 4, sumDigits, 3);
    if (sumDigits[0] != sum0
        || sumDigits[1] != sum1
        || sumDigits[2] != sum2)
    {
        fprintf(stderr,
                "update_spin_word: inconsistency with sumSpinWords function");
    }
    #endif
    
    /* We now have the number of "up" neighbours stored in the bits
        of sum0, sum1, and sum2: for a given bit position,
        we have the following possiblities:
        sum0	sum1	sum2	# of "up" neighbours
        --------------------------------------------
        0	0	0		0
        1	0	0		1
        0	1	0		2
        1	1	0		3
        0	0	1		4
        Note that if either of sum0 or sum1 is set, sum2 cannot be. */
    
    SpinWord new_spin = 0;
    mask = 1;
    for (bit = 0; bit < WORD_SIZE; bit++)
    {
        BoltzFactor	factor;
        
        if ( sum2 & mask )
        {
            factor = boltz_up4;
        }
        else
        {
            if ( sum1 & mask )
            {
                if ( sum0 & mask )
                    factor = boltz_up3;
                else
                    factor = boltz_up2;
            }
            else
            {
                if ( sum0 & mask )
                    factor = boltz_up1;
                else
                    factor = boltz_up0;
            }
        }
        
        BoltzFactor randnum = RAND_NUM();
        if ( randnum > factor )
        {
            new_spin |= mask;
        }
        
        mask <<= 1;
    }
    
    *spin_ptr = new_spin;
}

/***************************************************************************
NAME:		update_spins( spins, boltz_up )
INPUT:		Spins           *spins		: ptr to spins structure
                BoltzFactor     boltz_up[5]	: probabilities of a "down" spin
                                                  given the # of "up" neighbours
OUTPUT:		---
RETURN:		void
DESCRIPTION:	Uses the heat bath algorithm to update the spin configuration
		by traversing the 2D array, updating each spin in place,
		and then setting the padding layer spins to match their
		opposite numbers in order to enforce periodic boundary
		conditions.
EXT REFERENCES: ---
EXT EFFECTS:	---
CREATION:	Aug 4/91

---------------------------------------------------------------------------*/
void	update_spins( Spins *spins, BoltzFactor boltz_up[5] )
{
    int         n_rows = spins->numRows;
    int         n_cols = spins->numCols;
    SpinWord    *spinWords = spins->spinWords;
    
    SpinWord	*spin_ptr;
    int		row, row_size;
    int		n_wcols = num_spinwords_per_row(n_cols);
    
    /* We proceed along the rows and columns of 'spins',
        updating the spins in place */
    
    row_size = n_wcols + 2;
    spin_ptr = &spinWords[n_wcols + 3]; /* the first real spin */
    
    for (row = 0; row < n_rows; row++)
    {
        int	wcol;
        
        for (wcol = 0; wcol < n_wcols; wcol++)
        {
            update_spin_word( spin_ptr, row_size, boltz_up );
            ++spin_ptr;
        }
        
        /* Skip over the right and left boundary layer spins */
        spin_ptr += 2;
    }
    
    update_boundary_layer( spinWords, n_rows, n_wcols );
}

/***************************************************************************
NAME:		count_up_spins( spin )
INPUT:		SpinWord	spin		: a spin word
                int             num_bits        : number of bits to count
                                                  (starting with lowest bits)
OUTPUT:		---
RETURN:		int				: number of "up" spins
DESCRIPTION:	Counts the number of "up" spins in the spin-word.
EXT REFERENCES: ---
EXT EFFECTS:	---
CREATION:	Dec 6/91

---------------------------------------------------------------------------*/
static	int	count_up_spins( SpinWord spin, int num_bits )
{
    int	count;
    
    /* Just for efficiency, we do a quick check for all spins up */
    if ( spin == AllBitsSet )
    {
        count = num_bits;
    }
    else
    {
        if (num_bits < WORD_SIZE)
        {
            spin &= LowBits[num_bits - 1];
        }

        if (WORD_SIZE == 32)
        {
            // This is derived from the 'precomputed_bitcount' routine
            // on the web page "Counting Number of On Bits in an Integer"
            // by Gurmeet Singh Manku
            // http://www-db.stanford.edu/~manku/bitcount/bitcount.html
            
            count = BitsInChar[(spin      ) & 0xFFu]
                 +  BitsInChar[(spin >>  8) & 0xFFu]
                 +  BitsInChar[(spin >> 16) & 0xFFu]
                 +  BitsInChar[(spin >> 24) & 0xFFu];
        }
        else
        {
            // If you get here, you need to implement the above
            // for your WORD_SIZE. It should be a straightforward extension
            // of the above.
            printf("Need to implement 'count_up_spins' function for WORD_SIZE = %u\n",
                   WORD_SIZE);
            exit(1);
        }

    }
    
    return count;
}

/***************************************************************************
NAME:		calc_ising_statistics( spins, ptr_magn, ptr_nn_corr )
INPUT:		Spins         *spins		: ptr to spins structure
OUTPUT:		double		*ptr_magn	: magnetization
                double		*ptr_nn_corr	: nearest-neighbour correlation
RETURN:		void
DESCRIPTION:	Calculates the following statistics for the current
		configuration of spins:
		- the magnetization
		- the nearest-neighbour correlation
EXT REFERENCES: ---
EXT EFFECTS:	---
CREATION:	Nov 18/91

---------------------------------------------------------------------------*/
void	calc_ising_statistics( const Spins *spins,
                               double *ptr_magn, double *ptr_nn_corr )
{
    int         n_rows = spins->numRows;
    int         n_cols = spins->numCols;
    SpinWord    *spinWords = spins->spinWords;
    int		num_spins = n_rows * n_cols;

    SpinWord	*spin_ptr;
    int         row, sum, sum_nn;
    double	magn, nn_corr;
    
    int	n_wcols = num_spinwords_per_row( n_cols );
    
    // If 'n_cols' is not a multiple of WORD_SIZE, we need to be careful
    // not to include the "extra" padding spins in the statistics
    
// NB: at the moment we aren't excluding these extra spins
// And I no longer understand the equation being use dto calculate the nn_corr
// - why is it sufficient to add up spins only in the same row
// - don't we need to look at the spins in the rows above & below?
    
    sum = 0;
    sum_nn = 0;
    
    spin_ptr = &spinWords[n_wcols + 3]; /* the first real spin */
    
    for (row = 0; row < n_rows; row++)
    {
        int wcol;
        for (wcol = 0; wcol < n_wcols; wcol++)
        {
            sum += count_up_spins( spin_ptr[0], WORD_SIZE );
            sum_nn += count_up_spins( spin_ptr[0] & spin_ptr[1], WORD_SIZE );
            
            ++spin_ptr;
        }
        
        /* Skip over the right and left boundary layer spins */
        spin_ptr += 2;
    }
    
    magn = (2.0 * sum / num_spins) - 1.0;
    nn_corr = (4.0 * sum_nn / num_spins) - 2.0 * magn - 1.0;
    
    *ptr_magn = magn;
    *ptr_nn_corr = nn_corr;
}

/***************************************************************************
NAME:		calc_boltz( temperature, field, boltz_up )
INPUT:		double	temperature		: temperature
                double	field			: magnetic field
OUTPUT:		BoltzFactor boltz_up[5]	: array of Boltzmann factors
RETURN:		void
DESCRIPTION:	Calculates the Boltzmann factors for the current 'temperature'
		and 'field' and stores them in the 'boltz_up' array.
EXT REFERENCES: ---
EXT EFFECTS:	---
CREATION:	May 7/02	- extracted from set_ising_temperature()

---------------------------------------------------------------------------*/
void	calc_boltz( double temperature, double field, BoltzFactor boltz_up[5] )
{
    //printf("calc_boltz: temperature = %g field = %g\n",
    //                        temperature, field);
    
    for (int i = 0; i < 5; i++)
    {
        double	factor, e_up, e_down;
        
        factor = (2.0 * i - 4.0 + field) / temperature;
        // avoid floating point overflow/underflow by limiting factor
        if (factor < -100)
        {
            factor = -100;
        }
        else if (factor > 100)
        {
            factor = 100;
        }

        e_up = exp( -factor );
        e_down = exp( factor );

        //printf("calc_boltz: i = %d factor = %g e_up = %g e_down = %g\n",
        //                    i, factor, e_up, e_down);
        
#if SCALE_BOLTZ_TO_MAX_RANDOM
        boltz_up[i] = MAX_RANDOM * (e_up / (e_up + e_down));
        //printf("calc_boltz: boltz_up[%d] = %lu\n",
        //                    i, boltz_up[i]);
#else
        boltz_up[i] = e_up / (e_up + e_down);
#endif
    }
}

/***************************************************************************
NAME:		calc_block_spins_size( origSize, blockSize )
INPUT:		IsingSize   origSize        : size of original spins
                IsingSize   blockSize       : size of one block
OUTPUT:		---
RETURN:		IsingSize                   : size of block spins
DESCRIPTION:	Calculates the size of the Wilson block spins
EXT REFERENCES: ---
EXT EFFECTS:	---
CREATION:	April 21/05

---------------------------------------------------------------------------*/
IsingSize calc_block_spins_size(IsingSize origSize, IsingSize blockSize)
{
    // note that the division is integer division so it truncates
    IsingSize size = makeIsingSize(origSize.width  / blockSize.width,
                                   origSize.height / blockSize.height);
    return size;
}

/***************************************************************************
NAME:		calc_block_spins( origSpins, blockHeight, blockWidth )
INPUT:		Spins     *origSpins     : original spins
                int         blockHeight    : number of rows per block
                int         blockWidth     : number of columns per block
OUTPUT:		---
RETURN:		Spins *               : block spins
DESCRIPTION:	Calculates the Wilson block spins
                The caller is responsible for freeing the block spins.
EXT REFERENCES: ---
EXT EFFECTS:	---
CREATION:	April 14/05

---------------------------------------------------------------------------*/
Spins * calc_block_spins(Spins *origSpins, IsingSize blockSize)
{
    int         origNumRows  = origSpins->numRows;
    int         origNumCols  = origSpins->numCols;
    SpinWord    *origSpinWords = origSpins->spinWords;
    int         origNumWcols = num_spinwords_per_row(origNumCols);
    
    // we use the 'offsets' array as an optimization
    // to avoid doing modulus operations in the inner loop
    int         numOffsets = origNumWcols + blockSize.width;
    int         *offsets = myMalloc(numOffsets * sizeof(int));
    for (int i = 0; i < numOffsets; i++)
    {
        offsets[i] = i % origNumWcols;
    }
    
    // similarly, the 'origBitIndices' array is an optimization
    // to avoid doing division operations in the inner loop
    int         *origBitIndices = myMalloc(blockSize.width * sizeof(int));

    IsingSize   size = calc_block_spins_size(makeIsingSize(origNumCols, origNumRows),
                                             blockSize);
    int         numRows = size.height;
    int         numCols = size.width;
    Spins       *spins = allocate_spins(numRows, numCols);
    SpinWord    *spinWords = spins->spinWords;
    int         numWcols = num_spinwords_per_row(numCols);
    
    int numSpinsPerBlock = blockSize.width * blockSize.height;
    int halfUp = numSpinsPerBlock / 2;
    int needToDoTieBreaking = ((numSpinsPerBlock % 2) == 0);
    
    // this is not a very efficient way of doing it, but we'll fix that later
    for (int row = 0, origRow = 0; row < numRows; row++, (origRow += blockSize.height))
    {
        int origWordRowIndex = origNumWcols + 3
                                + origRow * (origNumWcols + 2);
        for (int col = 0, origCol = 0; col < numCols; col++, (origCol += blockSize.width))
        {
            int numUp = 0;
            #ifdef CHECK_CALC
            int numUp2 = 0;
            #endif
            int origWordBaseIndex = origWordRowIndex;
            int offsetBase = origCol % origNumWcols;
            
            // pre-calculate the origBitIndices as a small optimization
            for (int bCol = 0; bCol < blockSize.width; bCol++)
            {
                origBitIndices[bCol] = (origCol + bCol) / origNumWcols;
            }
            
            for (int bRow = 0; bRow < blockSize.height; bRow++, (origWordBaseIndex += (origNumWcols + 2)))
            {
                for (int bCol = 0; bCol < blockSize.width; bCol++)
                {
                    int origWordIndex = origWordBaseIndex + offsets[offsetBase + bCol];
                    SpinWord    origSpinWord = origSpinWords[origWordIndex];
                    numUp += ((origSpinWord & (1 << origBitIndices[bCol])) != 0);
                    
                    #ifdef CHECK_CALC
                    // The following line is an easier to understand
                    // but less efficient way to compute the number of up spins in the block:
                    numUp2 += spinIsUp(origSpins, origRow + bRow, origCol + bCol);
                    if (numUp2 != numUp)
                    {
                        fprintf(stderr,
                                "calc_block_spins: numUp: %d numUp2: %d",
                                numUp, numUp2);
                    }
                    #endif
                }
            }
            
            int blockIsUp;
            if (needToDoTieBreaking && numUp == halfUp)
            {
                blockIsUp = (RAND_NUM() > HALF_MAX);
            }
            else
            {
                blockIsUp = (numUp > halfUp);
            }
            
            int         wordIndex = numWcols + 3           // first real spin-word
                                    + row * (numWcols + 2) // which row
                                    + (col % numWcols);    // which word in that row
            int         bitIndex = col / numWcols;         // which bit in that word
            SpinWord  mask = (1 << bitIndex);
            
            if (blockIsUp)
            {
                spinWords[wordIndex] |= mask;
            }
            else
            {
                spinWords[wordIndex] &= (~mask);
            }
        }
    }
    
    myFree(offsets);
    myFree(origBitIndices);
    
    return spins;
}

/***************************************************************************
NAME:		spinIsUp( spins, row, col )
INPUT:		Spins       *spins          : spins
                int         row             : row index of the spin
                int         col             : column index of the spin
OUTPUT:		---
RETURN:		int                        : 1 if the spin is "up", 0 if "down"
DESCRIPTION:	Determines if one individual spin is "up" or "down".
                This is just for reference purposes and debugging
                - it is not usually efficient enough for general use.
EXT REFERENCES: ---
EXT EFFECTS:	---
CREATION:	Apr 15/05

---------------------------------------------------------------------------*/
int    spinIsUp(Spins *spins, int row, int col)
{
    SpinWord    *spinWords = spins->spinWords;
    int         numCols  = spins->numCols;
    int         numWcols = num_spinwords_per_row(numCols);

    int         wordIndex = numWcols + 3           // first real spin-word
                            + row * (numWcols + 2) // which row
                            + (col % numWcols);    // which word in that row
    int         bitIndex = col / numWcols;         // which bit in that word
    SpinWord    spinWord = spinWords[wordIndex];
    int         isUp = ((spinWord & (1 << bitIndex)) != 0);
    
    return isUp;
}

/***************************************************************************
NAME:		copy_spins( dest_spins, src_spins )
INPUT:		Spins     *src_spins	: source spins
OUTPUT:		Spins     *dest_spins	: destination spins
RETURN:		void
DESCRIPTION:	Copies 'srcSpins' into 'destSpins'
                The destination spins structure is modified to have the
                same number of row & columns as the source spins
                and the 'spinWords' array is reallocated if necessary.
EXT REFERENCES: ---
EXT EFFECTS:	---
CREATION:	June 4/02

---------------------------------------------------------------------------*/
void copy_spins( Spins *dest_spins, const Spins *src_spins)
{
    int         numRows = src_spins->numRows;
    int         numCols = src_spins->numCols;
    SpinWord    *src_spinWords = src_spins->spinWords;

    int	totNumWords = total_num_spinwords(numRows, numCols);
    
    if (dest_spins->numRows != numRows || dest_spins->numCols != numCols)
    {
        //fprintf(stderr, "copy_spins resizing the array (dest %d %d  src %d %d)\n",
        //        dest_spins->numRows, dest_spins->numCols, numRows, numCols);

        myFree(dest_spins->spinWords);
        dest_spins->numRows = numRows;
        dest_spins->numCols = numCols;
        dest_spins->spinWords
                  = (SpinWord *) myMalloc(totNumWords * sizeof(SpinWord));
    }
    
    memcpy(dest_spins->spinWords,
           src_spinWords,
           totNumWords * sizeof(SpinWord));
}

/***************************************************************************
NAME:		allocate_spins( n_rows, n_cols )
INPUT:		int	n_rows			: # of rows of spins
		int	n_cols			: # of columns of spins
						  (usually a multiple of WORD_SIZE)
OUTPUT:		---
RETURN:		Spins *			: ptr to spins structure
DESCRIPTION:	Allocates sufficient memory space for the n_rows x n_cols of spins
		plus the boundary spins. Returns a pointer to a Spins structure
                that has been initialized to have a 'spinWords' array
                of the correct size.
EXT REFERENCES: ---
EXT EFFECTS:	---
CREATION:	Oct 5/02

---------------------------------------------------------------------------*/
Spins *allocate_spins( int n_rows, int n_cols )
{
    int         totNumWords = total_num_spinwords(n_rows, n_cols);
    SpinWord    *spinWords = (SpinWord *) myMalloc(totNumWords * sizeof(SpinWord));
    
    Spins       *spins = (Spins *) myMalloc(sizeof(Spins));
    spins->spinWords = spinWords;
    spins->numRows = n_rows;
    spins->numCols = n_cols;
    
    return spins;
}


/***************************************************************************
NAME:		num_overflow_spins( n_cols )
INPUT:		int	n_cols		: # of columns of spins
                                          (usually a multiple of WORD_SIZE)
OUTPUT:		---
RETURN:		int			: # of spin_words per row
DESCRIPTION:	Calculates the number of "overflow" spins due to 'n_cols'
                not being a multiple of WORD_SIZE.
                E.g. if WORD_SIZE is 32 and n_cols is 37, n_overflow = 5
EXT REFERENCES: ---
EXT EFFECTS:	---
CREATION:	Apr 14/05

---------------------------------------------------------------------------*/
inline int num_overflow_spins( int n_cols )
{
    int n_overflow = (n_cols % WORD_SIZE);
    
    return n_overflow;
}

/***************************************************************************
NAME:		num_spinwords_per_row( n_cols )
INPUT:		int	n_cols		: # of columns of spins
                                          (usually a multiple of WORD_SIZE)
OUTPUT:		---
RETURN:		int			: # of spin_words per row
DESCRIPTION:	Calculates the number of spin_words per row.
                (not including the 2 boundary spin_words)
EXT REFERENCES: ---
EXT EFFECTS:	---
CREATION:	Apr 14/05

---------------------------------------------------------------------------*/
inline int num_spinwords_per_row( int n_cols )
{
    // There are nWcols spin_words plus 2 boundary spin_words in each row
    // (see diagram above)
    int n_overflow = num_overflow_spins( n_cols );
    int nWcols = (n_cols / WORD_SIZE) + (n_overflow ? 1 : 0);
    
    return nWcols;
}

/***************************************************************************
NAME:		total_num_spinwords( n_rows, n_cols )
INPUT:		int	n_rows		: # of rows of spins
                int	n_cols		: # of columns of spins
                                          (usually a multiple of WORD_SIZE)
OUTPUT:		---
RETURN:		int			: total # of spin_words
DESCRIPTION:	Calculates the total number of spin_words needed to hold
                the n_rows x n_cols of spins plus the boundary spins.
EXT REFERENCES: ---
EXT EFFECTS:	---
CREATION:	Apr 14/05

---------------------------------------------------------------------------*/
int total_num_spinwords( int n_rows, int n_cols )
{
    // There are nWcols spin_words plus 2 boundary spin_words in each row
    // and there is a boundary row at top and bottom (see diagram above)
    int         nWcols = num_spinwords_per_row(n_cols);
    int         totNumWords = (n_rows + 2) * (nWcols + 2);

    return totNumWords;
}

/***************************************************************************
NAME:		free_spins( spins )
INPUT:		Spins  *spins			: ptr to spins structure
OUTPUT:		---
RETURN:		void
DESCRIPTION:	Frees the memory space associated with the spins
EXT REFERENCES: ---
EXT EFFECTS:	---
CREATION:	Mar 9/05

---------------------------------------------------------------------------*/
void free_spins( Spins *spins )
{
    if (spins != NULL)
    {
        myFree(spins->spinWords);
        myFree(spins);
    }
}

/***************************************************************************
NAME:		makeIsingSize( width, height )
INPUT:		int width, height		: dimensions of spins
OUTPUT:		---
RETURN:		IsingSize
DESCRIPTION:	Creates an IsingSpin structure with 'width' & 'height'
EXT REFERENCES: ---
EXT EFFECTS:	---
CREATION:	Apr 21/05

---------------------------------------------------------------------------*/
IsingSize makeIsingSize(int width, int height)
{
    IsingSize size;
    size.width = width;
    size.height = height;
    return size;
}

/**END*OF*FILE*************************************************************/
